@@ -4,7 +4,7 @@ from abc import ABC, abstractmethod
 from dataclasses import dataclass
 import random
 from typing import (
-    Callable, Type, Any, List, Sequence, Tuple, Generator, Union, Optional
+    Callable, Type, Any, List, Tuple, Generator, Union, Optional, Sequence
 )
 
 from .utils import ProgressBar
@@ -73,8 +73,8 @@ class StreamedFunctionBiDirectional(StreamedFunction, ABC):
         self.predecessor = predecessor
         self.successor   = successor
 
-    def backward(self, stream: Stream, **kwargs) -> Stream:
-        raise NotImplementedError()
+    # def backward(self, stream: Stream, **kwargs) -> Stream:
+    #     raise NotImplementedError()
 
     def forward(self, stream: Stream, **kwargs) -> Stream:
         raise NotImplementedError()
@@ -85,16 +85,18 @@ class StreamedFunctionBiDirectional(StreamedFunction, ABC):
         return self.predecessor.root()
 
     def summary(self) -> List[Tuple[Optional[str], str]]:
-        # TODO: support summary for join
-        R = self.root()
-        N = R
-        nodes = [(R.name, R.__class__.__name__)]
-        while (N := N.successor) is not None:
-            nodes.append((N.name, N.__class__.__name__))
-        return nodes
+        root = self.root()
+        if self == root:
+            curr = root
+            nodes = [(root.name, root.__class__.__name__)]
+            while (curr := curr.successor) is not None:
+                nodes.append((curr.name, curr.__class__.__name__))
+            return nodes
+        else:
+            return root.summary()
 
 
-class ProcessingNodeSequential(StreamedFunctionBiDirectional):
+class ProcessingNodeSequential(StreamedFunctionBiDirectional, ABC):
     def __init__(
         self,
         function: Callable,
@@ -116,15 +118,18 @@ class ProcessingNodeSequential(StreamedFunctionBiDirectional):
             return self.successor(stream)
         return stream
 
+
 @dataclass
-class EmptyConfig():
+class EmptyConfig:
     pass
 
+
 @dataclass
-class Config():
+class Config:
     path: Union[List[str], str]
 
-class Dataset(ProcessingNodeSequential):
+
+class Dataset(ProcessingNodeSequential, ABC):
     def __init__(
         self,
         function: Callable,
@@ -220,16 +225,16 @@ class Dataset(ProcessingNodeSequential):
     def clone(self) -> Dataset:
         raise NotImplementedError()
 
-    def replicate(self) -> DatasetNode:
+    def replicate(self) -> Dataset:
         raise NotImplementedError()
 
     def add_successor(self, successor_class: Type[DatasetNode], **kwargs) -> DatasetNode:
         if not issubclass(successor_class, DatasetNode):
             raise TypeError(f"{__class__.__name__}: Successor class must be a Dataset subclass "
-                    f"definition, not {type(successor_class)}.")
+                            f"definition, not {type(successor_class)}.")
 
         new_node = successor_class(predecessor=self.clone(), successor=None, meta=self.meta, **kwargs)
-        new_node.predecessor.successor = new_node # type: ignore # predecessor is me!
+        new_node.predecessor.successor = new_node  # type: ignore # predecessor is me!
         return new_node
 
     def filter(self, f: Callable[..., bool], name: str = 'filter') -> DatasetNode:
@@ -247,23 +252,30 @@ class Dataset(ProcessingNodeSequential):
     def flat_map(self, f: Callable = lambda x: x, name: str = 'flat_map') -> DatasetNode:
         return self.add_successor(FlatMap, function=f, name=name)
 
-    def aggregate(self, init: Callable, agg: Callable, reduce: Callable = lambda x: x, name: str = 'aggregate') -> DatasetNode:
+    def aggregate(self, init: Callable, agg: Callable, reduce: Callable = lambda x: x,
+        name: str = 'aggregate'
+    ) -> DatasetNode:
         return self.add_successor(Aggregate, init=init, agg=agg, reduce=reduce, name=name)
 
     def limit(self, n: int, name: str = 'limit') -> DatasetNode:
         return self.add_successor(Limit, n=n, name=name)
 
     def join(self, other: Dataset,
-        key_left: Optional[Callable] = None, key_right: Optional[Callable] = None, key: Optional[Callable] = None,
+        key_left: Optional[Callable] = None, key_right: Optional[Callable] = None,
+        key: Optional[Callable] = None, type: str = "outer",
         name: str = 'join'
     ) -> JoinDatasetSource:
         # TODO: add support for name in JoinDatasetSource
-        return JoinDatasetSource(x=self, y=other, key_x=key_left, key_y=key_right, key=key, config=EmptyConfig())
+        return JoinDatasetSource(x=self, y=other, key_x=key_left, key_y=key_right, key=key, type=type, config=EmptyConfig())
 
     def zip(self, *iterables, name: str = 'zip') -> ZipDatasetSource:
         # TODO: add support for name in ZipDatasetSource
         return ZipDatasetSource(self, *iterables, config=EmptyConfig())
 
+    def cache(self, name: str = 'cache') -> CacheDatasetSource:
+        # TODO: add support for name in CacheDatasetSource
+        return CacheDatasetSource(source=self, config=EmptyConfig())
+
 
 class DatasetSource(Dataset):
     def __init__(
@@ -369,98 +381,15 @@ class DatasetSource(Dataset):
     def clone(self) -> DatasetSource:
         return self.replicate()
 
-    def replicate(self):
+    def replicate(self) -> DatasetSource:
         return self.__class__(config=self.config, successor=None)
 
     def _generate_examples(self) -> Generator:
         raise NotImplementedError()
 
 
-class JoinDatasetSource(DatasetSource):
-    def __init__(
-        self,
-        config: Any,
-        x: Dataset,
-        y: Dataset,
-        key_x: Optional[Callable] = None,
-        key_y: Optional[Callable] = None,
-        key: Optional[Callable] = None,
-        **kwargs
-    ):
-        super().__init__(config=config, **kwargs)
-        self.x = x.clone()
-        self.y = y.clone()
-        if not isinstance(x, Dataset):
-            raise TypeError("JoinDatasetSource: Argument x is not a Dataset")
-        if not isinstance(y, Dataset):
-            raise TypeError("JoinDatasetSource: Argument y is not a Dataset")
-        if key is not None and (key_x is not None or key_y is not None):
-            raise ValueError("JoinDatasetSource: either key or (key_x and key_y) must be specified")
-        if key is not None:
-            if not isinstance(key, Callable):
-                raise TypeError(f"JoinDatasetSource: Argument key is not a function")
-        else:
-            if not isinstance(key_x, Callable):
-                raise TypeError(f"JoinDatasetSource: Argument key_x is not a function")
-            if not isinstance(key_y, Callable):
-                raise TypeError(f"JoinDatasetSource: Argument key_y is not a function")
-        if key is not None:
-            self.key_x = key
-            self.key_y = key
-        else:
-            self.key_x = key_x
-            self.key_y = key_y
-        self.joined = {}
-
-    def dataset_name(self):
-        return "join"
-
-    def _generate_examples(self) -> Generator[dict, None, None]:
-        for data in self.x:
-            self._join(data, self.key_x)
-        for data in self.y:
-            self._join(data, self.key_y)
-        joined_data = self.joined
-        self.joined = None
-        yield joined_data
-
-    def _join(self, data: Any, key_fn: Callable):
-        key = key_fn(data)
-        if key not in self.joined:
-            self.joined[key] = []
-        self.joined[key].append(data)
-
-
-class ZipDatasetSource(DatasetSource):
-    def __init__(
-        self,
-        *iterables,
-        config: Any,
-        **kwargs
-    ):
-        super().__init__(config=config, **kwargs)
-        if not all(isinstance(ds, Dataset) for ds in iterables):
-            raise TypeError("ZipDatasetSource: One of the given iterables is not a Dataset")
-        self.iterables = iterables
-
-    def dataset_name(self):
-        return "zip"
-
-    def replicate(self) -> ZipDatasetSource:
-        return ZipDatasetSource(
-            *self.iterables,
-            config=self.config,
-            successor=self.successor,  # type: ignore
-            meta=self.meta,
-        )
-
-    def _generate_examples(self) -> Generator[tuple, None, None]:
-        for data_tuple in zip(*self.iterables):
-            yield data_tuple
-
-
 @dataclass
-class DatasetNode(Dataset):
+class DatasetNode(Dataset, ABC):
     def __init__(
         self,
         function: Callable,
@@ -557,6 +486,213 @@ class DatasetNode(Dataset):
         return replica
 
 
+class JoinDatasetSource(DatasetSource):
+    def __init__(
+        self,
+        config: Any,
+        x: Dataset,
+        y: Dataset,
+        type: str,
+        key_x: Optional[Callable] = None,
+        key_y: Optional[Callable] = None,
+        key: Optional[Callable] = None,
+        **kwargs
+    ):
+        super().__init__(config=config, **kwargs)
+        self.x = x.clone()
+        self.y = y.clone()
+        if not isinstance(x, Dataset):
+            raise TypeError("JoinDatasetSource: Argument x is not a Dataset")
+        if not isinstance(y, Dataset):
+            raise TypeError("JoinDatasetSource: Argument y is not a Dataset")
+        _available_type = ['inner', 'full', 'outer', 'left', 'right']
+        if type not in _available_type:
+            raise ValueError(f"JoinDatasetSource: unknown value for type: {type}. Known: {_available_type} (full=outer)")
+        if key is not None and (key_x is not None or key_y is not None):
+            raise ValueError("JoinDatasetSource: either key or (key_x and key_y) must be specified")
+        if key is not None:
+            if not isinstance(key, Callable):
+                raise TypeError("JoinDatasetSource: Argument key is not a function")
+        else:
+            if not isinstance(key_x, Callable):
+                raise TypeError("JoinDatasetSource: Argument key_x is not a function")
+            if not isinstance(key_y, Callable):
+                raise TypeError("JoinDatasetSource: Argument key_y is not a function")
+        if key is not None:
+            self.key_x = key
+            self.key_y = key
+        else:
+            self.key_x = key_x
+            self.key_y = key_y
+        self.type = type
+        # self.joined = {}
+
+    def dataset_name(self):
+        return "join"
+
+    def summary(self) -> List[Tuple[Optional[str], str]]:
+        join_sums = super().summary()
+        sources_sums = tuple((ds.summary() for ds in [self.x, self.y]))
+        join_sums[0] = join_sums[0] + sources_sums
+        return join_sums
+
+    def _generate_examples(self) -> Generator[dict, None, None]:
+        # for data in self.x:
+        #     self._join(data, self.key_x)
+        # for data in self.y:
+        #     self._join(data, self.key_y)
+        if self.type == "inner":
+            joined_data = self._join_inner()
+        elif self.type in ["outer", "full"]:
+            joined_data = self._join_outer()
+        elif self.type == "left":
+            joined_data = self._join_left()
+        elif self.type == "right":
+            joined_data = self._join_right()
+
+        # joined_data = self.joined
+        # self.joined = {}
+        yield joined_data
+
+    def _join_outer(self):
+        joined = {}
+        for data in self.x:
+            joined = self._append_anytime(joined, data, self.key_x)
+        for data in self.y:
+            joined = self._append_anytime(joined, data, self.key_y)
+        return joined
+
+    def _join_left(self):
+        joined = {}
+        for data in self.x:
+            joined = self._append_anytime(joined, data, self.key_x)
+        for data in self.y:
+            joined = self._append_if_exist(joined, data, self.key_y)
+        return joined
+
+    def _join_right(self):
+        joined = {}
+        for data in self.y:
+            joined = self._append_anytime(joined, data, self.key_y)
+        for data in self.x:
+            joined = self._append_if_exist(joined, data, self.key_x)
+        return joined
+
+    def _join_inner(self):
+        x_data = {}
+        for data in self.x:
+            x_data = self._append_anytime(x_data, data, self.key_x)
+        joined = {}
+        for data in self.y:
+            # joined = self.intersect(joined, data, self.key_y)
+            key = self.key_y(data)
+            if key in x_data:
+                joined[key] = x_data[key]
+                joined[key].append(data)
+        return joined
+
+    @staticmethod
+    def _append_if_exist(joined: dict, data: Any, key_fn: Callable):
+        key = key_fn(data)
+        if key in joined:
+            joined[key].append(data)
+        return joined
+
+    @staticmethod
+    def _append_anytime(joined: dict, data: Any, key_fn: Callable):
+        key = key_fn(data)
+        if key not in joined:
+            joined[key] = []
+        joined[key].append(data)
+        return joined
+
+    def replicate(self) -> JoinDatasetSource:
+        return JoinDatasetSource(
+            x=self.x,
+            y=self.y,
+            key_x=self.key_x,
+            key_y=self.key_y,
+            config=self.config,
+            type=self.type,
+            successor=self.successor,  # type: ignore
+            meta=self.meta,
+        )
+
+
+class ZipDatasetSource(DatasetSource):
+    def __init__(
+        self,
+        *iterables,
+        config: Any,
+        **kwargs
+    ):
+        super().__init__(config=config, **kwargs)
+        if not all(isinstance(ds, Dataset) for ds in iterables):
+            raise TypeError("ZipDatasetSource: One of the given iterables is not a Dataset")
+        self.iterables = iterables
+
+    def dataset_name(self):
+        return "zip"
+
+    def summary(self) -> List[Tuple[Optional[str], str]]:
+        zip_sums = super().summary()
+        iterable_sums = tuple((ds.summary() for ds in self.iterables))
+        zip_sums[0] = zip_sums[0] + iterable_sums
+        return zip_sums
+
+    def replicate(self) -> ZipDatasetSource:
+        return ZipDatasetSource(
+            *self.iterables,
+            config=self.config,
+            successor=self.successor,  # type: ignore
+            meta=self.meta,
+        )
+
+    def _generate_examples(self) -> Generator[tuple, None, None]:
+        for data_tuple in zip(*self.iterables):
+            yield data_tuple
+
+
+class CacheDatasetSource(DatasetSource):
+    def __init__(
+        self,
+        source: Dataset,
+        **kwargs
+    ):
+        super().__init__(**kwargs)
+        self.source = source
+        self.cached_data = []
+        self.finished = False
+
+    def dataset_name(self):
+        return "cache"
+
+    def summary(self) -> List[Tuple[Optional[str], str]]:
+        return self.source.summary() + super().summary()
+
+    def replicate(self) -> CacheDatasetSource:
+        return CacheDatasetSource(
+            source=self.source,
+            config=self.config,
+            successor=self.successor,  # type: ignore
+            meta=self.meta,
+        )
+
+    def _generate_examples(self) -> Generator[tuple, None, None]:
+        # Restart called after a keyboard interrupt in the middle of a generation
+        if self.cached_data != [] and self.finished == False:
+            self.cached_data = []
+
+        if self.cached_data == []:
+            for el in self.source:
+                self.cached_data.append(el)
+                yield el
+            self.finished = True
+        else:
+            for el in self.cached_data:
+                yield el
+
+
 @dataclass
 class Filter(DatasetNode):
     def __init__(
@@ -573,7 +709,7 @@ class Filter(DatasetNode):
             predecessor=predecessor,
             successor=successor,
             meta=meta,
-            name=name, 
+            name=name,
             **kwargs
         )
 
@@ -751,14 +887,14 @@ class Map(DatasetNode):
             raise TypeError(f"Map: Argument function must be a callable, not {type(function)}")
 
         super().__init__(
-            function=function, 
+            function=function,
             predecessor=predecessor,
             successor=successor,
             meta=meta,
-            name=name, 
+            name=name,
             **kwargs
         )
-    
+
     def compute(self, stream: Stream, **kwargs) -> Stream:
         output = self.function(stream.data)
         return self.forward(Stream(data=output))
@@ -788,14 +924,14 @@ class FlatMap(DatasetNode):
             raise TypeError(f"FlatMap: Argument function must be a callable, not {type(function)}")
 
         super().__init__(
-            function=function, 
+            function=function,
             predecessor=predecessor, # type: ignore
             successor=successor, # type: ignore
             meta=meta,
-            name=name, 
+            name=name,
             **kwargs
         )
-    
+
     def compute(self, stream: Stream, **kwargs) -> Generator[Stream, None, Union[PullStream, None]]:
         output = self.function(stream.data)
 
